@@ -7,10 +7,12 @@
 #include <stdbool.h>
 #include <string.h>
 
-#include "cafebabe/code_attribute.h"
-#include "cafebabe/class.h"
 #include "cafebabe/line_number_table_attribute.h"
+#include "cafebabe/stack_map_table_attribute.h"
+#include "cafebabe/exceptions_attribute.h"
+#include "cafebabe/code_attribute.h"
 #include "cafebabe/method_info.h"
+#include "cafebabe/class.h"
 
 #include "jit/compilation-unit.h"
 #include "jit/compiler.h"
@@ -51,29 +53,55 @@ struct vm_method {
 
 	struct cafebabe_code_attribute code_attribute;
 	struct cafebabe_line_number_table_attribute line_number_table_attribute;
+	struct cafebabe_stack_map_table_attribute stack_map_table_attribute;
+	struct cafebabe_exceptions_attribute exceptions_attribute;
 
 	struct compilation_unit *compilation_unit;
 	struct jit_trampoline *trampoline;
 
-	bool is_vm_native;
+	char flags;
+
+	unsigned int nr_annotations;
+	struct vm_annotation **annotations;
+	bool annotation_initialized;
 };
 
+#define VM_METHOD_FLAG_VM_NATIVE	(1 << 1)
+#define VM_METHOD_FLAG_TRACE		(1 << 2)
+#define VM_METHOD_FLAG_TRACE_GATE	(1 << 3)
+
+struct vm_method *vm_method_get_overridden(struct vm_method *vmm);
+
+int vm_method_do_init(struct vm_method *vmm);
+
 int vm_method_init(struct vm_method *vmm,
 	struct vm_class *vmc, unsigned int method_index);
 
 int vm_method_init_from_interface(struct vm_method *vmm, struct vm_class *vmc,
 	unsigned int method_index, struct vm_method *interface_method);
 
+int vm_method_init_annotation(struct vm_method *vmm);
+
 static inline bool vm_method_is_public(struct vm_method *vmm)
 {
 	return vmm->method->access_flags & CAFEBABE_METHOD_ACC_PUBLIC;
 }
 
+static inline bool vm_method_is_private(struct vm_method *vmm)
+{
+	return vmm->method->access_flags & CAFEBABE_METHOD_ACC_PRIVATE;
+}
+
 static inline bool vm_method_is_static(struct vm_method *vmm)
 {
 	return vmm->method->access_flags & CAFEBABE_METHOD_ACC_STATIC;
 }
 
+static inline bool vm_method_is_final(struct vm_method *vmm)
+{
+	return vmm->method->access_flags & CAFEBABE_METHOD_ACC_FINAL;
+}
+
 static inline bool vm_method_is_native(struct vm_method *vmm)
 {
 	return vmm->method->access_flags & CAFEBABE_METHOD_ACC_NATIVE;
@@ -111,13 +139,23 @@ static inline bool method_is_virtual(struct vm_method *vmm)
 static inline bool vm_method_is_jni(struct vm_method *vmm)
 {
 	return vmm->method->access_flags & CAFEBABE_METHOD_ACC_NATIVE
-		&& !vmm->is_vm_native;
+		&& !(vmm->flags & VM_METHOD_FLAG_VM_NATIVE);
 }
 
 static inline bool vm_method_is_vm_native(struct vm_method *vmm)
 {
 	return vmm->method->access_flags & CAFEBABE_METHOD_ACC_NATIVE
-		&& vmm->is_vm_native;
+		&& (vmm->flags & VM_METHOD_FLAG_VM_NATIVE);
+}
+
+static inline bool vm_method_is_traceable(struct vm_method *vmm)
+{
+	return vmm->flags & VM_METHOD_FLAG_TRACE;
+}
+
+static inline bool vm_method_is_trace_gate(struct vm_method *vmm)
+{
+	return vmm->flags & VM_METHOD_FLAG_TRACE_GATE;
 }
 
 static inline bool vm_method_is_special(struct vm_method *vmm)
@@ -125,6 +163,11 @@ static inline bool vm_method_is_special(struct vm_method *vmm)
 	return vmm->name[0] == '<';
 }
 
+static inline bool vm_method_is_compiled(struct vm_method *vmm)
+{
+	return compilation_unit_is_compiled(vmm->compilation_unit);
+}
+
 static inline enum vm_type method_return_type(struct vm_method *method)
 {
 	char *return_type = index(method->type, ')') + 1;
@@ -133,9 +176,13 @@ static inline enum vm_type method_return_type(struct vm_method *method)
 
 int vm_method_prepare_jit(struct vm_method *vmm);
 
-static inline void *vm_method_native_ptr(struct vm_method *vmm)
+static inline void *vm_method_entry_point(struct vm_method *vmm)
 {
-	return vmm->compilation_unit->native_ptr;
+	return cu_entry_point(vmm->compilation_unit);
+}
+static inline void *vm_method_ic_entry_point(struct vm_method *vmm)
+{
+	return cu_ic_entry_point(vmm->compilation_unit);
 }
 
 static inline void *vm_method_trampoline_ptr(struct vm_method *vmm)
@@ -143,24 +190,21 @@ static inline void *vm_method_trampoline_ptr(struct vm_method *vmm)
 	return buffer_ptr(vmm->trampoline->objcode);
 }
 
-/*
- * XXX: It is not safe to use this function during compilation unless
- * we are sure that @vmm is not currently being compiled.
- */
 static inline void *vm_method_call_ptr(struct vm_method *vmm)
 {
 	void *result;
 
-	pthread_mutex_lock(&vmm->compilation_unit->mutex);
-
-	if (vmm->compilation_unit->is_compiled)
-		result = vm_method_native_ptr(vmm);
+	if (vm_method_is_compiled(vmm))
+		result = vm_method_entry_point(vmm);
 	else
 		result = vm_method_trampoline_ptr(vmm);
 
-	pthread_mutex_unlock(&vmm->compilation_unit->mutex);
-
 	return result;
 }
 
+static inline bool vm_method_is_missing(struct vm_method *vmm)
+{
+	return !vmm->compilation_unit;
+}
+
 #endif
